
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

const int loops = 12; 

struct bufferCntrl{ 
    unsigned int size; 
    unsigned int freeSlots;
    pthread_mutex_t bufferMutex;
    pthread_cond_t  bufferNotFull; 
    pthread_cond_t  bufferNotEmpty; 
};

typedef struct bufferCntrl BUFFER_CNTRL; 


int initBufferCntrl(BUFFER_CNTRL *bufferCntrl, int size_p){
    bufferCntrl->size = size_p;
    bufferCntrl->freeSlots = size_p;
    pthread_mutex_init(&bufferCntrl->bufferMutex, NULL);
    pthread_cond_init(&bufferCntrl->bufferNotFull, NULL);
    pthread_cond_init(&bufferCntrl->bufferNotEmpty, NULL);
    return 0; 
}

int freeBufferCntrl(BUFFER_CNTRL *bufferCntrl){
    pthread_mutex_destroy(&bufferCntrl->bufferMutex);
    pthread_cond_destroy(&bufferCntrl->bufferNotFull);
    pthread_cond_destroy(&bufferCntrl->bufferNotEmpty);
    return 0; 
}
 

void *produce(void *bufferCntrl){
    BUFFER_CNTRL *bctl = (BUFFER_CNTRL *)bufferCntrl; 
    pthread_mutex_t *bufferMutexP = &bctl->bufferMutex;
    pthread_cond_t  *bufferNotFullP  = &bctl->bufferNotFull;
    pthread_cond_t  *bufferNotEmptyP  = &bctl->bufferNotEmpty;
    unsigned int    *sizeP        = &bctl->size;
    unsigned int    *freeSlotsP   = &bctl->freeSlots;
    int i;

    sleep(2);
    printf("\n producer ptr: %p, %p, %p", 
             bufferMutexP, bufferNotFullP, bufferNotEmptyP); 
    printf("\n producer started, freeSlots: %d", *freeSlotsP);


    for(i = 0; i < loops; i++){
        /* produce item */ 
	printf("\n producer requires mutex, i: %2d", i); 
	fflush(stdout);
	pthread_mutex_lock(bufferMutexP);
	printf("\n producer got mutex, free: %d of %d", *freeSlotsP, *sizeP); 
        if (*freeSlotsP < 1){
	    printf("\n producer going to wait"); 
            fflush(stdout);
	    pthread_cond_wait(bufferNotFullP, bufferMutexP); 
	}
        (*freeSlotsP)--;
        printf("\n producer made an item"); 
        sleep(1);
        if(*freeSlotsP == *sizeP - 1){
	    printf("\n producer signalling buffer not empty"); 
            fflush(stdout);
	    pthread_cond_signal(bufferNotEmptyP); 
	}
	printf("\n producer unlocking mutex"); 
	fflush(stdout);
	pthread_mutex_unlock(bufferMutexP);
        sleep(1);
    }
    return NULL;
}
   
void *consume(void *bufferCntrl){
    BUFFER_CNTRL *bctl = (BUFFER_CNTRL *)bufferCntrl; 
    pthread_mutex_t *bufferMutexP = &bctl->bufferMutex;
    pthread_cond_t  *bufferNotFullP  = &bctl->bufferNotFull;
    pthread_cond_t  *bufferNotEmptyP  = &bctl->bufferNotEmpty;
    unsigned int    *sizeP        = &bctl->size;
    unsigned int    *freeSlotsP   = &bctl->freeSlots;
    int i;

    printf("\n consumer ptr: %p, %p, %p", 
             bufferMutexP, bufferNotFullP, bufferNotEmptyP); 
    printf("\n consumer started, freeSlots: %d", *freeSlotsP);
    sleep(1);

    for(i = 0; i < loops; i++){
	printf("\n consumer requires mutex, i: %2d ", i); 
	fflush(stdout);
	pthread_mutex_lock(bufferMutexP); 
	printf("\n consumer got mutex, free: %d of %d", *freeSlotsP, *sizeP); 
        if (*freeSlotsP >= *sizeP){
	    printf("\n consumer going to wait"); 
            fflush(stdout);
	    pthread_cond_wait(bufferNotEmptyP, bufferMutexP); 
	}
        (*freeSlotsP)++;
        printf("\n    consumer got an item"); 
        sleep(1);

        if(*freeSlotsP == 1){
	    printf("\n consumer signalling buffer not full");
            fflush(stdout);
	    pthread_cond_signal(bufferNotFullP); 
	}
	printf("\n consumer unlocking mutex"); 
	fflush(stdout);
	pthread_mutex_unlock(bufferMutexP);
        sleep(3);
    }
    return NULL;
}
   

int main(int argc, char *argv[]){
    int result; 
    pthread_t producer;  
    pthread_t consumer;  
    pthread_attr_t pt_attr;

    BUFFER_CNTRL bufferCntrl; 
   
    initBufferCntrl(&bufferCntrl, 5);

    
    pthread_attr_init(&pt_attr);
    pthread_attr_setdetachstate(&pt_attr, PTHREAD_CREATE_JOINABLE);

    result = pthread_create(&producer, &pt_attr, produce, &bufferCntrl); 
    if (result != 0){
	printf("\n error creating thread producer: %d", result);
    }
    printf("\n created producer thread");
    result = pthread_create(&consumer, &pt_attr, consume, &bufferCntrl); 
    if (result != 0){
	printf("\n error creating thread producer: %d", result);
    }
    printf("\n created consumer thread\n");
    fflush(stdout); 
    pthread_join(producer, NULL);
    pthread_join(consumer, NULL);
    printf("\n producer and consumer terminated ");
    freeBufferCntrl(&bufferCntrl); 
    printf("\n mutex and condition vars released \n");
    fflush(stdout);     
    pthread_exit(NULL);
    return 0; 
}



